/*	kanji.c -
 *	"Convert JIS kanji string to postscript using TeX font"
 *	
 *	original from "dvi2ps"
 *	modified by Kenjiro Yamanaka & Kazuhiro Kazama
 *	NTT Software Lab.
 */

/*
 *	external definitions
 */
#include <sys/param.h>
#include <signal.h>
#include <stdio.h>
#include <ctype.h>
#include <sys/file.h>
/*#include <malloc.h>*/
#include <strings.h>
#include "defs.h"
#include "font.h"

#ifndef FONTAREA
#define  FONTAREA         "/usr/lib/tex/fonts"
#endif
  
/*
 *	Global Procedures
 */
FILE*	BINARYOPEN();
void	GetBytes();
void	GetFontDef();
int     HasBeenRead();
int     NoSignExtend();         /* see cautionary note in code, re arithmetic vs logical shifts */
void	OpenFontFile();
void    PutInt();
int	ReadFontDef();
void	SetChar();
void	SetFntNum();
int     SignExtend();           /* see cautionary note in code, re arithmetic vs logical shifts */

void    PutPsChar();
void    ku2tex();  

/*
 *	Global Variables
 */
extern char **fontname;

#ifdef DEBUG
int Debug = 0;
#endif

int	hconv = 31576;

struct font_entry *fontptr;	/* font_entry pointer                       */
struct font_entry *hfontptr=NULL; /* font_entry pointer                     */
	
int   nopen;			/* number of open FNT files                 */
struct font_entry *pfontptr = NULL; /* previous font_entry pointer          */
struct openfont_list openfont_files[MAXOPEN+1];	/* list of open FNT file identifiers        */
int openfont_time = 0;
	
void PutKanjiPsChar(c1, c2, f_size)
unsigned char	c1, c2;
float	f_size;
{
    unsigned char	f, n;

    ku2tex(c1, c2, &f, &n);
    PutPsChar(f, n, f_size);
}

void PutPsChar(f_n,n,f_size)
unsigned char f_n,n;
float f_size;
{
    if (!HasBeenRead(f_n))
	ReadFontDef(f_n, f_size);
    SetFntNum(f_n,TRUE);
    SetChar(n,f_size);
}

/*
 *	ku2tex() -
 *	KUTEN code to (font_number,char_number)
 */
void	ku2tex(uin, din, uout, dout)
unsigned char	uin, din, *uout, *dout;
{
    int	i, j, k;
	
    i = (int)JIS2KU(uin);
    j = (int)JIS2KU(din);
    
    if (i == 0)			/* sy */
    {
	*uout = SY;
	*dout = j + 1;
    }
    else if (i == 1)		/* sy */
    {
	*uout = SY;
	*dout = j + 101;
    }
    else if (i == 2)		/* roma */
    {
	*uout = ROMA;
	*dout = j + 33;
    }
    else if (i == 3)		/* hira */
    {
	*uout = HIRA;
	*dout = j + 1;
    }
    else if (i == 4)		/* kata */
    {
	*uout = KATA;
	*dout = j + 1;
    }
    else if (i == 5)		/* greek */
    {
	*uout=GREEK;
	*dout=j+1;
    }
    else if (i == 6)		/* russian */
    {
	*uout = RUSSIAN;
	*dout = j + 1;
    }
    else if (i == 7)		/* keisen */
    {
	*uout = KEISEN;
	*dout = j+1;
    }
    else if (15 <= i && i < 47)	/* DAI 1 */
    {
	i -= 15;
	k = (NKU * i) + j;
	*uout = (k / 256) + DAI1;
	*dout = (k % 256);
    }
    else if (47 <= i && i < 84)	/* DAI 2 */
    {
	i -= 47;
	k = (NKU * i) + j;
	*uout = (k / 256) + DAI2;
	*dout = (k % 256);
    }
    else
    {
	fprintf(stderr,"%d %d isn't a JIS char\n",uin, din);
	exit(1);
    }
}

/*
 *	ReadFontDef
 */
int	ReadFontDef(k, f_size)
int	k;
float	f_size;
{
    register int i;
    register struct font_entry *tfontptr;/* temporary font_entry pointer   */
    register struct char_entry *tcharptr;/* temporary char_entry pointer  */

    switch ((int)f_size)
    {
    case 5:
    case 6:
    case 7:
    case 8:
    case 9:
    case 10:
    case 12:
    case 17:
    case 20:
	break;
    default:
	fprintf(stderr, "no font(size %d)\n", (int)f_size);
	exit(1);
    }
    
    if ((tfontptr = NEW(struct font_entry)) == NULL)
    {
	fprintf(stderr, "can't malloc space for font_entry\n");
	exit(1);
    }
    tfontptr->next = hfontptr;
    tfontptr->font_file_id = NULL;
    fontptr = hfontptr = tfontptr;
    
    tfontptr->ncdl = 0;
    
    tfontptr->a = 0;
    tfontptr->k = k;
    tfontptr->s = SPACESIZE;
    tfontptr->s = DESIGNSIZE;
    sprintf(tfontptr->n,"%s%d",fontname[k], (int)f_size);
    tfontptr->font_mag = (float)RESOLUTION;
    
    /* default for FNT files (allow missing FNT files) */
    tfontptr->font_type = FT_PXL;
    tfontptr->dic_pack = FALSE;
    tfontptr->magnification = 0;
    tfontptr->designsize = 0;
    tfontptr->nfntchars = NFNTCHARS;
    for (i = FIRSTFNTCHAR; i <= LASTFNTCHAR; i++)
    {
	tcharptr = &(tfontptr->ch[i]);
	tcharptr->width = 0;
	tcharptr->height = 0;
	tcharptr->xOffset= 0;
	tcharptr->yOffset = 0;
	tcharptr->psfont = NONFONT;
	tcharptr->pschar = NONCHAR;
	tcharptr->where.fileOffset = NONEXISTANT;
	tcharptr->tfmw = 0;
    }
    if (SearchFontFile(tfontptr, FONTAREA))
	readpsfontinfo(tfontptr);
}	

/*
 *	HasBeenRead
 */
int	HasBeenRead(k)
int	k;
{
    struct font_entry *ptr;
    
    ptr = hfontptr;
    while ((ptr!=NULL) && (ptr->k!=k))
	ptr = ptr->next;
    return( ptr != NULL );
}

/*
 *	OpenFontFile
 */
void	OpenFontFile(fe)
register struct font_entry *fe;
{
    register int i,least_used,current;
    int prio;
    FILE *fntfp;
	
#ifdef DEBUG
    if (Debug) printf("Open Font file\n");
#endif
    openfont_time++;
    if (pfontptr == fe)
	return;                 /* we need not have been called */
    
    for (current = 1;
	 (current <= nopen) &&
	 (openfont_files[current].openfont_file_id != fe->font_file_id);
	 ++current)
	;                       /* try to find file in open list */
    
    if (current <= nopen)       /* file already open */
    {
	if( (fntfp = openfont_files[current].openfont_file_id) != NO_FILE )
	    fseek(fntfp,0,0);	/* reposition to start of file */
	}	
	else                        /* file not in open list */
	{
	    if (nopen < MAXOPEN)    /* just add it to list */
		current = ++nopen;
	    else                    /* list full -- find least used file, */
	    {                       /* close it, and reuse slot for new file */
		least_used = 1;
		prio = priority(least_used);
		for (i = 2; i <= MAXOPEN; ++i)
		    if(prio > priority(i))
			prio = priority(least_used = i);
		if (openfont_files[least_used].openfont_file_id != NO_FILE) {
		    FILE *fid;
		    struct font_entry *fp;
		    
		    fid = openfont_files[least_used].openfont_file_id;
		    fp=hfontptr;    /* mark file as being closed in the entry */
		    while (fp!=NULL && fp->font_file_id != fid)
			fp = fp->next;
		    if (fp==NULL)
		    {
			fprintf(stderr, "Open file %x not found in font entry list.\n", fid);
			exit(1);
		    }
		    else
		    {
			fp->font_file_id = NULL;
		    }
		    fclose( fid );
		}
		current = least_used;
	    }
	    if ((fntfp=BINARYOPEN(fe->name,"r")) == NULL)
	    {
		fprintf(stderr, "FNT file %s could not be opened\n",fe->name);
		fntfp = NO_FILE;
	    }
	    openfont_files[current].openfont_file_id = fntfp;
	    openfont_files[current].use_count = 0;
	}
    pfontptr = fe;			/* make previous = current font */
    fe->font_file_id = fntfp;		/* set file identifier */
    openfont_files[current].use_count++;	/* update reference count */
    openfont_files[current].last_use = openfont_time;
}


priority(i)
int	i;
{
    return(openfont_files[i].use_count -
	   (openfont_time-openfont_files[i].last_use));
}

/*
 *	SetFntNum
 *  this routine is used to specify the font to be used in printing future
 *   characters
 */
void	SetFntNum(k, Emitting)
int	k, Emitting;
{
    fontptr = hfontptr;
    while ((fontptr!=NULL) && (fontptr->k!=k))
	fontptr = fontptr->next;
    if (fontptr == NULL)
    {
	fprintf(stderr, "font %d undefined\n", k);
	exit(1);
    }
}

/*
 *	SetChar
 */
void	SetChar(c,f_size)
int	c;
float	f_size;
{
    register struct char_entry *ce;  /* temporary char_entry pointer */
    
    psfontdict(fontptr, c);
    ce = &(fontptr->ch[c]);

    if(ce->psfont != NONFONT)	/* ignore missing fonts */
    {
/*	setpsfont(ce,f_size);*/
	setpsfont(ce);
	PutInt(ce->pschar);
	printf("jshow\n");
    }
}
